"""boto3 glue client utilities."""
import json
import logging
import time
import timeit
from typing import Any, Dict, List, Optional

import boto3
import s3path

from wpdatautil.pathlib import path_to_uri

log = logging.getLogger(__name__)

# Ref: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glue.html


def create_database(**kwargs: Any) -> None:
    """Safely create the specified AWS Glue database.

    At minimum the `Name` keyword argument is required.

    :param kwargs: These are forwarded to `create_database` as its `CatalogId` and `DatabaseInput` parameters.
    """
    name = kwargs["Name"]
    call_kwargs = {"DatabaseInput": {k: v for k, v in kwargs.items() if k != "CatalogId"}}
    assert call_kwargs
    if catalog_id := kwargs.get("CatalogId"):
        call_kwargs["CatalogId"] = catalog_id
    client = boto3.client("glue")

    try:
        response = client.create_database(**call_kwargs)
    except client.exceptions.AlreadyExistsException:
        log.info(f"AWS Glue has a database named {name}.")
    else:
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
        log.info(f"Created AWS Glue database {name}.")


def delete_database(**kwargs: Any) -> None:
    """Safely delete the specified AWS Glue database.

    At minimum the `Name` keyword argument is required.

    :param kwargs: These are forwarded to `delete_database`.
    """
    name = kwargs["Name"]
    call_kwargs = kwargs.copy()
    if not call_kwargs.get("CatalogId"):
        call_kwargs.pop("CatalogId", None)  # Removes key if having None value.
    client = boto3.client("glue")

    try:
        response = client.delete_database(**call_kwargs)
    except client.exceptions.EntityNotFoundException:
        log.info(f"AWS Glue does not have a database named {name}.")
    else:
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
        log.info(f"Deleted AWS Glue database {name}.")


def ensure_and_run_crawler(**kwargs: Any) -> List[dict]:
    """Return the list of database tables after ensuring and running the specified AWS Glue crawler with the given configuration.

    :param kwargs: These are indirectly forwarded to `create_crawler`.
    """
    crawler_name = kwargs["Name"]
    ensure_crawler(**kwargs)
    run_crawler(crawler_name)

    db_name = kwargs.get("DatabaseName") or crawler_name
    tables = get_tables(DatabaseName=db_name)
    table_names = [t["Name"] for t in tables]
    log.info(f"After crawling {crawler_name}, {len(tables)} tables exist: {', '.join(table_names)}")
    return tables


def ensure_and_run_s3_tables_crawler(crawler: str, s3_path: s3path.S3Path, s3_exclusions: Optional[List[str]] = None, **kwargs: Any) -> List[dict]:
    """Return the list of database tables after ensuring and running the specified AWS Glue crawler to crawl the top-level tables in the given S3 path.

    :param s3_exclusions: These are forwarded for each S3 path.
    :param kwargs: These are indirectly forwarded to `create_crawler`.
    """
    # Ref: https://stackoverflow.com/a/67185483/
    assert all(k not in kwargs for k in ("Name", "Targets"))

    s3_crawl_extra_kwargs = {"Exclusions": s3_exclusions} if s3_exclusions else {}
    s3_crawl_targets: List[Dict[str, Any]] = [
        {"Path": path_to_uri(p), **s3_crawl_extra_kwargs} for p in s3_path.glob("*")
    ]  # For correct crawling, each table's path is independently included.
    assert s3_crawl_targets
    log.info(f"{len(s3_crawl_targets)} S3 paths are to be crawled, the first of which is {s3_crawl_targets[0]['Path']}.")

    tables = ensure_and_run_crawler(
        Name=crawler, Targets={"S3Targets": s3_crawl_targets}, Configuration=json.dumps({"Version": 1.0, "Grouping": {"TableGroupingPolicy": "CombineCompatibleSchemas"}}), **kwargs
    )
    return tables


def ensure_crawler(**kwargs: Any) -> None:
    """Ensure that the specified AWS Glue crawler exists with the given configuration.

    At minimum the `Name` and `Targets` keyword arguments are required. Various default arguments are used.

    :param kwargs: These are forwarded to `create_crawler`.
    """
    # Use defaults
    assert all(kwargs.get(k) for k in ("Name", "Targets"))
    defaults = {
        "Role": "AWSGlueRole",
        "DatabaseName": kwargs["Name"],
        "SchemaChangePolicy": {"UpdateBehavior": "UPDATE_IN_DATABASE", "DeleteBehavior": "DELETE_FROM_DATABASE"},
        "RecrawlPolicy": {"RecrawlBehavior": "CRAWL_EVERYTHING"},
        "LineageConfiguration": {"CrawlerLineageSettings": "DISABLE"},
    }
    kwargs = {**defaults, **kwargs}

    # Ensure crawler
    client = boto3.client("glue")
    name = kwargs["Name"]
    try:
        response = client.create_crawler(**kwargs)
        log.info(f"Created AWS Glue crawler {name}.")
    except client.exceptions.AlreadyExistsException:
        response = client.update_crawler(**kwargs)
        log.info(f"Updated AWS Glue crawler {name}.")
    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200


def get_tables(skip_temp: bool = False, **kwargs: Any) -> List[dict]:
    """Return the list of tables in the specified AWS Glue database.

    At minimum the `DatabaseName` keyword argument is required.

    :param skip_temp: If true, table names starting with "temp_" are skipped.
    :param kwargs: These are forwarded to `get_tables`.
    """
    assert "DatabaseName" in kwargs
    tables = boto3.client("glue").get_paginator("get_tables").paginate(**kwargs).build_full_result()["TableList"]
    if skip_temp:
        # Note: Such tables are sometimes automatically created.
        tables = [t for t in tables if not t["Name"].startswith("temp_")]
    return tables


def recreate_database(**kwargs: Any) -> None:
    """Safely delete and create the specified AWS Glue database.

    At minimum the `Name` keyword argument is required.

    :param kwargs: These are forwarded.
    The `CatalogId` (optional) and `Name` (required) keyword arguments are forwarded to `delete_database`.
    All kwargs are forwarded to `create_database` as its `CatalogId` and `DatabaseInput` parameters.
    """
    delete_database(CatalogId=kwargs.get("CatalogId"), Name=kwargs["Name"])  # Note: Additional kwargs, such as for `create_database`, are not relevant.
    create_database(**kwargs)


def run_crawler(crawler: str, *, timeout_minutes: int = 120, retry_seconds: int = 5) -> None:
    """Run the specified AWS Glue crawler, waiting until completion."""
    # Ref: https://stackoverflow.com/a/66072347/
    timeout_seconds = timeout_minutes * 60
    client = boto3.client("glue")
    start_time = timeit.default_timer()
    abort_time = start_time + timeout_seconds

    def wait_until_ready() -> None:
        state_previous = None
        while True:
            response_get = client.get_crawler(Name=crawler)
            state = response_get["Crawler"]["State"]
            if state != state_previous:
                log.info(f"Crawler {crawler} is {state.lower()}.")
                state_previous = state
            if state == "READY":  # Other known states: RUNNING, STOPPING
                return
            if timeit.default_timer() > abort_time:
                raise TimeoutError(f"Failed to crawl {crawler}. The allocated time of {timeout_minutes:,} minutes has elapsed.")
            time.sleep(retry_seconds)

    wait_until_ready()
    response_start = client.start_crawler(Name=crawler)
    assert response_start["ResponseMetadata"]["HTTPStatusCode"] == 200
    log.info(f"Crawling {crawler}.")
    wait_until_ready()
    log.info(f"Crawled {crawler}.")



"""Presto SQL utilities."""
from typing import List


def date_parse(formats: List[str], /, column: str = "{col}") -> str:
    """Return the coalesced date_parse usage string to parse a column using the given formats.

    Example:
    -------
        date_parse(['%m/%d/%Y %r', '%Y-%m-%d %H:%i:%s', '%Y-%m-%d %H:%i:%s.%f']) ->
            "coalesce(try(date_parse({col}, '%m/%d/%Y %r')), try(date_parse({col}, '%Y-%m-%d %H:%i:%s')), date_parse({col}, '%Y-%m-%d %H:%i:%s.%f'))"

    """
    assert formats

    def _parse(fmt: str, /) -> str:
        return f"date_parse({column}, '{fmt}')"

    if len(formats) == 1:
        return _parse(formats[0])

    return "coalesce(" + ", ".join(f"try({_parse(f)})" for f in formats[:-1]) + ", " + _parse(formats[-1]) + ")"



"""awswrangler athena utilities."""
import logging

import awswrangler as wr

from wpdatautil.object import fullname
from wpdatautil.timeit import Timer

log = logging.getLogger(__name__)


def run_query(sql: str, database: str, description: str = "query") -> None:
    """Run the given AWS Athena query, waiting until completion."""
    timer = Timer()
    log.info(f"Submitted the {description}.")
    try:
        wr.athena.wait_query(wr.athena.start_query_execution(sql=sql, database=database))
    except Exception as exception:
        log.error(f"Failed to complete the {description}: {fullname(exception)}: {exception}")
        raise
    log.info(f"Completed in {timer} the {description}.")



"""str SQL utilities."""
import functools
import textwrap
from typing import Dict, List, Optional

from wpdatautil.textwrap import indent4


class SqlQuery:
    """SQL query helper to apply transformations to input columns, returning the output columns and query."""

    def __init__(self, inputs: List[str], *, transformations: Dict[str, Optional[str]], template: str, indent: int = 0) -> None:
        """Return a SQL query object.

        :param inputs: Input column names.
        :param transformations: Applicable transformations to input columns. If a column is to be transformed, its name is the key and its value is the transformation.
                                In the value, {col} is replaced with the respective column name. If the value is None, the column is skipped.
        :param template: SQL query to format. In it, {selections} is formatted with the actual selections.
        :param indent: The output query is indented by this number of spaces.

        Example: SqlQuery(['c1', 'c2', 'c3'],
                  transformations={'c2': None, 'c3': '{col} * 2', 'c4': 'count(*)'},
                  template="SELECT {selections}
                            FROM t")
        """
        self._inputs, self._transformations, self._template, self._indent = inputs, transformations, template, indent

    @functools.cached_property
    def selections(self) -> Dict[str, str]:
        """Return the output column selections.

        Example: {'c1': 'c1', 'c3': 'c3 * 2', 'c4': 'count(*)'}
        """
        return sql_column_selections(self._inputs, transformations=self._transformations)

    @functools.cached_property
    def outputs(self) -> List[str]:
        """Return the output column names.

        Example: ['c1', 'c3', 'c4']
        """
        return list(self.selections)

    @functools.cached_property
    def query(self) -> str:
        """Return the formatted query.

        Example:
        -------
            SELECT
                c1,
                c3 * 2 AS c3,
                count(*) AS c4
            FROM t

        """
        query = self._template.format(selections=sql_column_selections_str(self.selections))
        query = textwrap.indent(query, prefix=" " * self._indent)
        return query


def sql_column_selections(inputs: List[str], *, transformations: Dict[str, Optional[str]]) -> Dict[str, str]:
    """Return a mapping of SQL SELECT clauses given the input columns after applying the given transformations.

    Example: sql_column_selections(['c1', 'c2', 'c3'], transformations={'c2': None, 'c3': '{col} * 2', 'c4': 'count(*)'}) -> {'c1': 'c1', 'c3': 'c3 * 2', 'c4': 'count(*)'}

    :param inputs: Input column names.
    :param transformations: Applicable transformations. If a column is to be transformed, its name is the key and its value is the transformation.
                            In the value, {col} is replaced with the respective column name. If the value is None, the column is skipped.
    """
    outputs = inputs + [c for c in transformations if c not in inputs]
    del inputs
    selections = {}

    for column_name in outputs:
        if column_name in transformations:
            transformed_column = transformations[column_name]
            if transformed_column is None:
                continue  # Skip column.
            transformed_column = transformed_column.format(col=column_name)
            selections[column_name] = transformed_column
        else:
            selections[column_name] = column_name

    return selections


def sql_column_selections_str(inputs: Dict[str, str]) -> str:
    """Return a string representation of a mapping of SQL SELECT clauses.

    Example: {'c1': 'c1', 'c2': 'c2 * 2'} -> 'c1, c2 * 2 AS c2'
    """
    selections = ", ".join(k if k == v else f"\n{v} AS {k}" for k, v in inputs.items())
    if "\n" in selections:
        selections = "\n" + indent4(selections, levels=1)
    return selections
